use alt6bfast::{Graph, AuxGraph};
use std::io::stdin;
use std::collections::HashSet;
use std::time::Instant;

fn parse_input() -> Vec<Vec<u8>> {
	return stdin().lines().map(|line| line.unwrap().into_bytes()).collect();
}

const NORTH: usize = 0;
const EAST: usize = 1;
const SOUTH: usize = 2;
const WEST: usize = 3;

/* Simulate the procedure and return a vector of coordinate pairs of all positions. */
fn simulate(grid: &Vec<Vec<u8>>, mut r: usize, mut c: usize) -> HashSet<(usize, usize)> {
	let mut dr: isize = -1;
	let mut dc: isize = 0; // start pointing up
	let nr = grid.len();
	let nc = (&grid[0]).len();
	let mut positions: HashSet<(usize, usize)> = HashSet::new();
	loop {
		// Get next position
		// println!("{r}, {c}");
		let r2: usize = r.wrapping_add_signed(dr);
		let c2: usize = c.wrapping_add_signed(dc);
		// If we're about to walk out of range, then r2 will be either nr or usize::MAX
		// == two's complement equivalent of -1
		if r2 >= nr || c2 >= nc {
			return positions;
		}
		// If we're about to walk into an obstacle, turn.
		if grid[r2][c2] == b'#' {
			(dr, dc) = (dc, -dr);
		} else {
			// Otherwise, advance.
			(r, c) = (r2, c2);
		}
		positions.insert((r, c));
	}
}

fn main() {
	let mut start = Instant::now();
	let grid = parse_input();
	// Find initial position
	let nr = grid.len();
	let nc = (&grid[0]).len();
	let mut r_start: usize = usize::MAX;
	let mut c_start: usize = usize::MAX;
	for r2 in 0..nr {
		for c2 in 0..nc {
			if grid[r2][c2] == b'^' {
				r_start = r2;
				c_start = c2;
			}
		}
	}
	if r_start >= nr || c_start >= nc {
		panic!("Could not find starting position");
	}
	println!("Initial parsing: {0} us", start.elapsed().as_micros());
	start = Instant::now();
	let graph = Graph::from_grid(&grid);
	println!("Edge list processing: {0} us", start.elapsed().as_micros());
	start = Instant::now();
	let graph = AuxGraph::from(graph);
	println!("Connected component processing: {0} us", start.elapsed().as_micros());
	start = Instant::now();
	let unobstructed_path: HashSet<(usize, usize)> = simulate(&grid, r_start, c_start);
	println!("Finding unobstructed path: {0} us", start.elapsed().as_micros());
	start =Instant::now();
	
	let mut answer = 0;
	let mut new_edge_starts = Vec::new();
	let mut new_edge_ends = Vec::new();
	let start_ix = 4 * (nc * r_start + c_start);
	let sink = 4 * nc * nr;
	let code = |r, c, dir| {4 * (r * nc + c) + dir};

	for (r, c) in unobstructed_path.into_iter() {
		// eprintln!("Adding obstacle at {r}, {c}");
		// We do a lot of recomputing of r*nc+c here, but the tradeoff for code clarity
		// is worth it. A good compiler could inline the calls to code and do common
		// subexpression elimination, in any case.
		
		// Add edges corresponding to a new obstacle in (r, c)
		if r != 0 {
			// cell above turns from looking south to west
			new_edge_starts.push(code(r-1, c, SOUTH));
			new_edge_ends.push(code(r-1, c, WEST));
		}
		if c != 0 {
			// cell to left turns from looking east to south
			new_edge_starts.push(code(r, c-1, EAST));
			new_edge_ends.push(code(r, c-1, SOUTH));
		}
		if r != nr-1 {
			// cell below turns from looking north to east
			new_edge_starts.push(code(r+1, c, NORTH));
			new_edge_ends.push(code(r+1, c, EAST));
		}
		if c != nc-1 {
			// cell to right turns from looking west to north
			new_edge_starts.push(code(r, c+1, WEST));
			new_edge_ends.push(code(r, c+1, NORTH));
		}
		if graph.adding_edges_creates_cycle(start_ix, &new_edge_starts, &new_edge_ends, sink) {
			// println!("{r}, {c}");
			answer += 1;
		}
		new_edge_starts.clear();
		new_edge_ends.clear();
	}
	println!("All queries: {0} us", start.elapsed().as_micros());
	println!("Answer: {answer}");
}
